'''
Created on 1 Sep. 2017

@author: jakubb
'''


import os.path
from nipype.interfaces.base import (
    TraitedSpec, BaseInterface, File, isdefined)
import nibabel as nib
import numpy as np


class CoreUmapCalcInputSpec(TraitedSpec):
    air__mask = File(mandatory=True, desc='air mask file')
    bones__mask = File(mandatory=True, desc='bone mask file')
    ute1_reg = File(mandatory=True, desc='ute echo 1 image file')
    ute2_reg = File(mandatory=True, desc='ute echo 2 image file')
    sute_cont_template = File(
        genfile=True,
        desc='sute continuous map in template space')
    sute_fix_template = File(
        genfile=True,
        desc='sute fixed map in template space')


class CoreUmapCalcOutputSpec(TraitedSpec):
    sute_cont_template = File(
        exists=True,
        desc='sute continuous map in template space')
    sute_fix_template = File(
        exists=True,
        desc='sute fixed map in template space')


class CoreUmapCalc(BaseInterface):
    """Creates two umaps in the template space"""

    input_spec = CoreUmapCalcInputSpec
    output_spec = CoreUmapCalcOutputSpec

    def _run_interface(self, runtime):
        air = nib.load(self.inputs.air__mask)
        bones = nib.load(self.inputs.bones__mask)
        ute1 = nib.load(self.inputs.ute1_reg)
        ute2 = nib.load(self.inputs.ute2_reg)

        r2star_map = 1000. * (np.log(np.array(ute1.get_data())) 
                              - np.log(np.array(ute2.get_data()))) / 2.39
        nans = np.isnan(r2star_map)
        r2star_map[nans] = 0

        u_bone = 0.000001351 * (r2star_map**3) - 0.003617 * (r2star_map**2) + 3.841 * r2star_map - 19.46
        # CONVERSION from HU to PET u values
        # Conversion based on:
        # Carney J P et al. (2006) Med. Phys. 33 976-83
        
        a = 0.000051
        b = 0.0471
        BP = 1047.

        low = u_bone < BP
        u_bone[low] = 0.000096 * (1000. + u_bone[low])
        high = u_bone >= BP
        u_bone[high] = a * (1000. + u_bone[high]) + b

        low_u = u_bone < 0.1134
        u_bone[low_u] = 0.1134
        # End of conversion
        
        u_soft_fixed = 0.1
        u_air = 0.

        umap = 10000. * (u_air * np.array(air.get_data()) +
                         u_bone * np.array(bones.get_data()) +
                         (1 - np.array(bones.get_data())) *
                         (1 - np.array(air.get_data())) * u_soft_fixed)

        nans = np.isnan(umap)
        umap[nans] = 0
        save_im = nib.Nifti1Image(umap, affine=ute1.affine)
        nib.save(save_im, self._gen_filename('sute_cont_template'))

        u_bone2 = 0.151
        umap2 = 10000. * (u_air * np.array(air.get_data()) +
                         u_bone2 * np.array(bones.get_data()) +
                         (1 - np.array(bones.get_data())) *
                         (1 - np.array(air.get_data())) * u_soft_fixed)

        nans = np.isnan(umap2)
        umap2[nans] = 0
        save_im = nib.Nifti1Image(umap2, affine=ute1.affine)
        nib.save(save_im, self._gen_filename('sute_fix_template'))

        return runtime

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs['sute_cont_template'] = self._gen_sute_cont_template_fname()
        outputs['sute_fix_template'] = self._gen_sute_fix_template_fname()
        return outputs

    def _gen_filename(self, name):
        if name == 'sute_cont_template':
            fname = self._gen_sute_cont_template_fname()
        elif name == 'sute_fix_template':
            fname = self._gen_sute_fix_template_fname()
        else:
            assert False
        return fname

    def _gen_sute_cont_template_fname(self):
        if isdefined(self.inputs.sute_cont_template):
            sute_cont_template_fname = self.inputs.sute_cont_template
        else:
            sute_cont_template_fname = os.path.join(
                os.getcwd(),
                "sute_cont_template.nii.gz")
        return sute_cont_template_fname

    def _gen_sute_fix_template_fname(self):
        if isdefined(self.inputs.sute_fix_template):
            sute_fix_template_fname = self.inputs.sute_fix_template
        else:
            sute_fix_template_fname = os.path.join(
                os.getcwd(),
                "sute_fix_template.nii.gz")
        return sute_fix_template_fname




from nipype.interfaces.base import (
    BaseInterface, BaseInterfaceInputSpec, TraitedSpec, Directory, File,
    traits)
import os
import shutil
import pydicom
import numpy as np
import glob


class PrepareFIXInputSpec(BaseInterfaceInputSpec):

    melodic_dir = Directory()
    filtered_epi = File(exists=True)
    t1_brain = File(exists=True)
    mc_par = File(exists=True)
    epi_brain_mask = File(exists=True)
    epi_preproc = File(exists=True)
    epi2t1_mat = File(exists=True)
    t12epi_mat = File(exists=True)
    t12MNI_mat = File(exists=True)
    MNI2t1_mat = File(exists=True)
    epi_mean = File(exists=True)


class PrepareFIXOutputSpec(TraitedSpec):

    fix_dir = Directory()
    hand_label_file = File(exists=True)


class PrepareFIX(BaseInterface):

    input_spec = PrepareFIXInputSpec
    output_spec = PrepareFIXOutputSpec

    def _run_interface(self, runtime):

        melodic_dir = self.inputs.melodic_dir
        filtered_epi = self.inputs.filtered_epi
        t1_brain = self.inputs.t1_brain
        mc_par = self.inputs.mc_par
        epi_brain_mask = self.inputs.epi_brain_mask
        epi_preproc = self.inputs.epi_preproc
        epi2t1_mat = self.inputs.epi2t1_mat
        t12epi_mat = self.inputs.t12epi_mat
        t12MNI_mat = self.inputs.t12MNI_mat
        MNI2t1_mat = self.inputs.MNI2t1_mat
        epi_mean = self.inputs.epi_mean

        shutil.copytree(melodic_dir, 'melodic_ica')
        os.mkdir('melodic_ica/reg')
        shutil.copy2(t12MNI_mat, 'melodic_ica/reg/highres2std.mat')
        shutil.copy2(MNI2t1_mat, 'melodic_ica/reg/std2highres.mat')
        shutil.copy2(epi2t1_mat, 'melodic_ica/reg/example_func2highres.mat')
        shutil.copy2(t1_brain, 'melodic_ica/reg/highres.nii.gz')
        shutil.copy2(epi_preproc, 'melodic_ica/reg/example_func.nii.gz')
        shutil.copy2(t12epi_mat, 'melodic_ica/reg/highres2example_func.mat')
        os.mkdir('melodic_ica/mc')
        shutil.copy2(mc_par, 'melodic_ica/mc/prefiltered_func_data_mcf.par')
        shutil.copy2(epi_brain_mask, 'melodic_ica/mask.nii.gz')
        shutil.copy2(epi_mean, 'melodic_ica/mean_func.nii.gz')
        shutil.copytree(melodic_dir, 'melodic_ica/filtered_func_data.ica')
        shutil.copy2(filtered_epi, 'melodic_ica/filtered_func_data.nii.gz')

        with open('hand_label_file.txt', 'w') as f:
            f.write('not_provided')

        return runtime

    def _list_outputs(self):
        outputs = self._outputs().get()

        outputs["fix_dir"] = os.getcwd()+'/melodic_ica'
        outputs["hand_label_file"] = os.getcwd()+'/hand_label_file.txt'

        return outputs


class FieldMapTimeInfoInputSpec(BaseInterfaceInputSpec):

    fm_mag = Directory()


class FieldMapTimeInfoOutputSpec(TraitedSpec):

    delta_te = traits.Float()


class FieldMapTimeInfo(BaseInterface):

    input_spec = FieldMapTimeInfoInputSpec
    output_spec = FieldMapTimeInfoOutputSpec

    def _run_interface(self, runtime):

        fm_mag = sorted(glob.glob(self.inputs.fm_mag+'/*'))
        tes = [pydicom.read_file(x).EchoTime for x in fm_mag]
        tes = list(set(tes))
        if len(tes) != 2:
            print('Something went wrong when trying to estimate '
                  'the delta TE between the two echos field map '
                  'images. delta_TE will be set equal to 2.46, '
                  'but please check if this is correct.')
            self.delta_te = 2.46
        else:
            self.delta_te = float(np.abs(tes[1]-tes[0]))
        print('delta_TE: {}'.format(self.delta_te))

        return runtime

    def _list_outputs(self):
        outputs = self._outputs().get()

        outputs["delta_te"] = self.delta_te

        return outputs



class BananaError(Exception):
    pass


class BananaUsageError(BananaError):
    pass


class BananaMissingHeaderValue(BananaError):
    pass


class BananaRuntimeError(BananaError):
    pass


class BananaTestSetupError(BananaError):
    pass



# Standard library imports
import os

# Third-party imports
import numpy as np

# Local imports
from nipype.interfaces.base import (OutputMultiPath, TraitedSpec, isdefined,
                                    traits, InputMultiPath, File)
from nipype.interfaces.spm.base import (SPMCommand, scans_for_fnames,
                                        SPMCommandInputSpec)
from nipype.utils.filemanip import split_filename


class MultiChannelSegmentInputSpec(SPMCommandInputSpec):
    channel_files = InputMultiPath(File(exists=True),
                                   desc="A list of files to be segmented",
                                   field='channel', copyfile=False,
                                   mandatory=True)
    channel_info = traits.Tuple(traits.Float(), traits.Float(),
                                traits.Tuple(traits.Bool, traits.Bool),
                                desc="""A tuple with the following fields:
            - bias reguralisation (0-10)
            - FWHM of Gaussian smoothness of bias
            - which maps to save (Corrected, Field) - a tuple of two boolean values""",  # @IgnorePep8
                                field='channel')
    tissues = traits.List(
        traits.Tuple(traits.Tuple(File(exists=True), traits.Int()),
                     traits.Int(), traits.Tuple(traits.Bool, traits.Bool),
                     traits.Tuple(traits.Bool, traits.Bool)),
        desc="""A list of tuples (one per tissue) with the following fields:
            - tissue probability map (4D), 1-based index to frame
            - number of gaussians
            - which maps to save [Native, DARTEL] - a tuple of two boolean values
            - which maps to save [Unmodulated, Modulated] - a tuple of two boolean values""",  # @IgnorePep8
        field='tissue')
    affine_regularization = traits.Enum('mni', 'eastern', 'subj', 'none',
                                        field='warp.affreg',
                                        desc='mni, eastern, subj, none ')
    warping_regularization = traits.Either(traits.List(traits.Float(),
                                                       minlen=5, maxlen=5),
                                           traits.Float(),
                                           field='warp.reg',
                                           desc=('Warping regularization '
                                                 'parameter(s). Accepts float '
                                                 'or list of floats (the '
                                                 'latter is required by '
                                                 'SPM12)'))
    sampling_distance = traits.Float(field='warp.samp',
                                     desc=('Sampling distance on data for '
                                           'parameter estimation'))
    write_deformation_fields = traits.List(traits.Bool(), minlen=2, maxlen=2,
                                           field='warp.write',
                                           desc=("Which deformation fields to "
                                                 "write:[Inverse, Forward]"))


class MultiChannelSegmentOutputSpec(TraitedSpec):
    native_class_images = traits.List(traits.List(File(exists=True)),
                                      desc='native space probability maps')
    dartel_input_images = traits.List(traits.List(File(exists=True)),
                                      desc='dartel imported class images')
    normalized_class_images = traits.List(traits.List(File(exists=True)),
                                          desc='normalized class images')
    modulated_class_images = traits.List(traits.List(File(exists=True)),
                                         desc=('modulated+normalized class '
                                               'images'))
    transformation_mat = OutputMultiPath(File(exists=True),
                                         desc='Normalization transformation')
    bias_corrected_images = OutputMultiPath(File(exists=True),
                                            desc='bias corrected images')
    bias_field_images = OutputMultiPath(File(exists=True),
                                        desc='bias field images')
    forward_deformation_field = OutputMultiPath(File(exists=True))
    inverse_deformation_field = OutputMultiPath(File(exists=True))


class MultiChannelSegment(SPMCommand):
    """Use spm_preproc8 (New Segment) to separate structural images into
    different tissue classes. Supports multiple modalities.

    NOTE: This interface currently supports single channel input only

    http://www.fil.ion.ucl.ac.uk/spm/doc/manual.pdf#page=43

    Examples
    --------
    >>> import nipype.interfaces.spm as spm
    >>> seg = spm.Segment()
    >>> seg.inputs.channel_files = 'structural.nii'
    >>> seg.inputs.channel_info = (0.0001, 60, (True, True))
    >>> seg.run() # doctest: +SKIP

    For VBM pre-processing [http://www.fil.ion.ucl.ac.uk/~john/misc/VBMclass10.pdf],
    TPM.nii should be replaced by /path/to/spm8/toolbox/Seg/TPM.nii

    >>> seg = Segment()
    >>> seg.inputs.channel_files = 'structural.nii'
    >>> tissue1 = (('TPM.nii', 1), 2, (True,True), (False, False))
    >>> tissue2 = (('TPM.nii', 2), 2, (True,True), (False, False))
    >>> tissue3 = (('TPM.nii', 3), 2, (True,False), (False, False))
    >>> tissue4 = (('TPM.nii', 4), 2, (False,False), (False, False))
    >>> tissue5 = (('TPM.nii', 5), 2, (False,False), (False, False))
    >>> seg.inputs.tissues = [tissue1, tissue2, tissue3, tissue4, tissue5]
    >>> seg.run() # doctest: +SKIP

    """

    input_spec = MultiChannelSegmentInputSpec
    output_spec = MultiChannelSegmentOutputSpec

    def __init__(self, **inputs):
        _local_version = SPMCommand().version
        if _local_version and '12.' in _local_version:
            self._jobtype = 'spatial'
            self._jobname = 'preproc'
        else:
            self._jobtype = 'tools'
            self._jobname = 'preproc8'

        SPMCommand.__init__(self, **inputs)

    def _format_arg(self, opt, spec, val):
        """Convert input to appropriate format for spm
        """

        if opt in ['channel_files', 'channel_info']:
            # structure have to be recreated because of some weird traits error
            new_channel = {}
            new_channel['vols'] = scans_for_fnames(self.inputs.channel_files)
            if isdefined(self.inputs.channel_info):
                info = self.inputs.channel_info
                new_channel['biasreg'] = info[0]
                new_channel['biasfwhm'] = info[1]
                new_channel['write'] = [int(info[2][0]), int(info[2][1])]
            return [new_channel]
        elif opt == 'tissues':
            new_tissues = []
            for tissue in val:
                new_tissue = {}
                new_tissue['tpm'] = np.array([','.join([tissue[0][0],
                                                        str(tissue[0][1])])],
                                             dtype=object)
                new_tissue['ngaus'] = tissue[1]
                new_tissue['native'] = [int(tissue[2][0]), int(tissue[2][1])]
                new_tissue['warped'] = [int(tissue[3][0]), int(tissue[3][1])]
                new_tissues.append(new_tissue)
            return new_tissues
        elif opt == 'write_deformation_fields':
            return super(MultiChannelSegment, self)._format_arg(opt, spec,
                                                                [int(val[0]),
                                                                 int(val[1])])
        else:
            return super(MultiChannelSegment, self)._format_arg(opt, spec, val)

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs['native_class_images'] = []
        outputs['dartel_input_images'] = []
        outputs['normalized_class_images'] = []
        outputs['modulated_class_images'] = []
        outputs['transformation_mat'] = []
        outputs['bias_corrected_images'] = []
        outputs['bias_field_images'] = []
        outputs['inverse_deformation_field'] = []
        outputs['forward_deformation_field'] = []

        n_classes = 5
        if isdefined(self.inputs.tissues):
            n_classes = len(self.inputs.tissues)
        for i in range(n_classes):
            outputs['native_class_images'].append([])
            outputs['dartel_input_images'].append([])
            outputs['normalized_class_images'].append([])
            outputs['modulated_class_images'].append([])

        for filename in self.inputs.channel_files:
            pth, base, ext = split_filename(filename)
            if isdefined(self.inputs.tissues):
                for i, tissue in enumerate(self.inputs.tissues):
                    if tissue[2][0]:
                        outputs['native_class_images'][i].append(
                            os.path.join(pth, "c%d%s.nii" % (i + 1, base)))
                    if tissue[2][1]:
                        outputs['dartel_input_images'][i].append(
                            os.path.join(pth, "rc%d%s.nii" % (i + 1, base)))
                    if tissue[3][0]:
                        outputs['normalized_class_images'][i].append(
                            os.path.join(pth, "wc%d%s.nii" % (i + 1, base)))
                    if tissue[3][1]:
                        outputs['modulated_class_images'][i].append(
                            os.path.join(pth, "mwc%d%s.nii" % (i + 1, base)))
            else:
                for i in range(n_classes):
                    outputs['native_class_images'][i].append(
                        os.path.join(pth, "c%d%s.nii" % (i + 1, base)))
            outputs['transformation_mat'].append(
                os.path.join(pth, "%s_seg8.mat" % base))

            if isdefined(self.inputs.write_deformation_fields):
                if self.inputs.write_deformation_fields[0]:
                    outputs['inverse_deformation_field'].append(
                        os.path.join(pth, "iy_%s.nii" % base))
                if self.inputs.write_deformation_fields[1]:
                    outputs['forward_deformation_field'].append(
                        os.path.join(pth, "y_%s.nii" % base))

            if isdefined(self.inputs.channel_info):
                if self.inputs.channel_info[2][0]:
                    outputs['bias_corrected_images'].append(
                        os.path.join(pth, "m%s.nii" % (base)))
                if self.inputs.channel_info[2][1]:
                    outputs['bias_field_images'].append(
                        os.path.join(pth, "BiasField_%s.nii" % (base)))
        return outputs



__version__ = '0.3.7'

__authors__ = [
    ("Thomas G. Close", "tom.g.close@gmail.com"),
    ("Francesco Sforazzini", "francesco.sforazzini@gmail.com"),
    ("Phillip G. D. Ward", "phillipgdward@gmail.com")]


install_requires = ['arcana>=0.4.5',
                    'pydicom>=1.0',
                    'nibabel>=2.3.0']



from nipype.interfaces.fsl.model import MELODIC
from nipype.interfaces.afni.preprocess import Volreg
from nipype.interfaces.fsl.utils import ImageMaths, ConvertXFM
from banana.interfaces.fsl import (FSLFIX, FSLFixTraining,
                                       SignalRegression, PrepareFIXTraining)
from arcana.data import FilesetSpec, InputFilesetSpec
from arcana.study.base import StudyMetaClass
from banana.requirement import (
    afni_req, fix_req, fsl_req, ants_req, c3d_req)
from banana.citation import fsl_cite
from banana.file_format import (
    nifti_gz_format, nifti_gz_x_format, rfile_format, directory_format,
    zip_format, par_format, text_format, dicom_format, text_matrix_format)
from banana.interfaces.afni import Tproject
from nipype.interfaces.utility import Merge as NiPypeMerge
import os.path as op
from nipype.interfaces.utility.base import IdentityInterface
from arcana.study import ParamSpec, SwitchSpec
from banana.study.mri.epi import EpiStudy
from nipype.interfaces.ants.resampling import ApplyTransforms
from banana.study.mri.t1 import T1Study
from arcana.study.multi import (
    MultiStudy, SubStudySpec, MultiStudyMetaClass)
from arcana.data import InputFileset
from arcana.utils.interfaces import CopyToDir
from nipype.interfaces.afni.preprocess import BlurToFWHM
from banana.interfaces.custom.bold import PrepareFIX
from banana.interfaces.c3d import ANTs2FSLMatrixConversion
import logging
from arcana.exceptions import ArcanaNameError
from banana.bids import BidsInput, BidsAssocInput

logger = logging.getLogger('banana')


atlas_path = op.abspath(
    op.join(op.dirname(__file__), '..', '..', '..', 'atlases'))

IMAGE_TYPE_TAG = ('0008', '0008')
PHASE_IMAGE_TYPE = ['ORIGINAL', 'PRIMARY', 'P', 'ND']
MAG_IMAGE_TYPE = ['ORIGINAL', 'PRIMARY', 'M', 'ND', 'NORM']


class BoldStudy(EpiStudy, metaclass=StudyMetaClass):

    add_data_specs = [
        InputFilesetSpec('train_data', rfile_format, optional=True,
                            frequency='per_study'),
        FilesetSpec('hand_label_noise', text_format,
                    'fix_preparation_pipeline'),
        FilesetSpec('labelled_components', text_format,
                    'fix_classification_pipeline'),
        FilesetSpec('cleaned_file', nifti_gz_format,
                    'fix_regression_pipeline'),
        FilesetSpec('filtered_data', nifti_gz_format,
                    'rsfMRI_filtering_pipeline'),
        FilesetSpec('mc_par', par_format, 'rsfMRI_filtering_pipeline'),
        FilesetSpec('melodic_ica', zip_format,
                    'single_subject_melodic_pipeline'),
        FilesetSpec('fix_dir', zip_format, 'fix_preparation_pipeline'),
        FilesetSpec('normalized_ts', nifti_gz_format,
                    'timeseries_normalization_to_atlas_pipeline'),
        FilesetSpec('smoothed_ts', nifti_gz_format,
                    'smoothing_pipeline')]

    add_param_specs = [
        ParamSpec('component_threshold', 20),
        ParamSpec('motion_reg', True),
        ParamSpec('highpass', 0.01),
        ParamSpec('brain_thresh_percent', 5),
        ParamSpec('MNI_template', op.join(atlas_path,
                                              'MNI152_T1_2mm.nii.gz')),
        ParamSpec('MNI_template_mask', op.join(
            atlas_path, 'MNI152_T1_2mm_brain_mask.nii.gz')),
        SwitchSpec('linear_reg_method', 'ants',
                   ('flirt', 'spm', 'ants', 'epireg')),
        ParamSpec('group_ica_components', 15)]

    primary_bids_selector = BidsInput(
        spec_name='magnitude', type='bold', format=nifti_gz_x_format)

    default_bids_inputs = [primary_bids_selector,
                           BidsAssocInput(
                               spec_name='field_map_phase',
                               primary=primary_bids_selector,
                               association='phasediff',
                               format=nifti_gz_format,
                               drop_if_missing=True),
                           BidsAssocInput(
                               spec_name='field_map_mag',
                               primary=primary_bids_selector,
                               association='phasediff',
                               type='magnitude',
                               format=nifti_gz_format,
                               drop_if_missing=True)]

    def rsfMRI_filtering_pipeline(self, **name_maps):

        pipeline = self.new_pipeline(
            name='rsfMRI_filtering',
            desc=("Spatial and temporal rsfMRI filtering"),
            citations=[fsl_cite],
            name_maps=name_maps)

        afni_mc = pipeline.add(
            'AFNI_MC',
            Volreg(
                zpad=1,
                out_file='rsfmri_mc.nii.gz',
                oned_file='prefiltered_func_data_mcf.par'),
            inputs={
                'in_file': ('preproc', nifti_gz_format)},
            wall_time=5,
            requirements=[afni_req.v('16.2.10')])

        filt = pipeline.add(
            'Tproject',
            Tproject(
                stopband=(0, 0.01),
                polort=3,
                blur=3,
                out_file='filtered_func_data.nii.gz'),
            inputs={
                'delta_t': ('tr', float),
                'mask': ('brain_mask', nifti_gz_format),
                'in_file': (afni_mc, 'out_file')},
            wall_time=5,
            requirements=[afni_req.v('16.2.10')])

        meanfunc = pipeline.add(
            'meanfunc',
            ImageMaths(op_string='-Tmean', suffix='_mean'),
            wall_time=5,
            inputs={
                'in_file': (afni_mc, 'out_file')},
            requirements=[fsl_req.v('5.0.10')])

        pipeline.add(
            'add_mean',
            ImageMaths(op_string='-add'),
            inputs={
                'in_file': (filt, 'out_file'),
                'in_file2': (meanfunc, 'out_file')},
            outputs={
                'filtered_data': ('out_file', nifti_gz_format),
                'mc_par': ('oned_file', par_format)},
            wall_time=5,
            requirements=[fsl_req.v('5.0.10')])

        return pipeline

    def single_subject_melodic_pipeline(self, **name_maps):

        pipeline = self.new_pipeline(
            name='MelodicL1',
            desc=("Single subject ICA analysis using FSL MELODIC."),
            citations=[fsl_cite],
            name_maps=name_maps)

        pipeline.add(
            'melodic_L1',
            MELODIC(
                no_bet=True,
                bg_threshold=self.parameter('brain_thresh_percent'),
                report=True,
                out_stats=True,
                mm_thresh=0.5,
                out_dir='melodic_ica'),
            inputs={
                'mask': ('brain_mask', nifti_gz_format),
                'tr_sec': ('tr', float),
                'in_files': ('filtered_data', nifti_gz_format)},
            outputs={
                'melodic_ica': ('out_dir', directory_format)},
            wall_time=15,
            requirements=[fsl_req.v('5.0.10')])

        return pipeline

    def fix_preparation_pipeline(self, **name_maps):

        pipeline = self.new_pipeline(
            name='prepare_fix',
            desc=("Pipeline to create the right folder structure before "
                  "running FIX"),
            citations=[fsl_cite],
            name_maps=name_maps)

        struct_ants2fsl = pipeline.add(
            'struct_ants2fsl',
            ANTs2FSLMatrixConversion(
                ras2fsl=True,
                reference_file=self.parameter('MNI_template')),
            inputs={
                'itk_file': ('coreg_to_template_mat', text_matrix_format),
                'source_file': ('coreg_ref_brain', nifti_gz_format)},
            requirements=[c3d_req.v('1.1.0')])
        epi_ants2fsl = pipeline.add(
            'epi_ants2fsl',
            ANTs2FSLMatrixConversion(
                ras2fsl=True),
            inputs={
                'source_file': ('brain', nifti_gz_format),
                'itk_file': ('coreg_matrix', text_matrix_format),
                'reference_file': ('coreg_ref_brain', nifti_gz_format)},
            requirements=[c3d_req.v('1.1.0')])

        MNI2t1 = pipeline.add(
            'MNI2t1',
            ConvertXFM(
                invert_xfm=True),
            inputs={
                'in_file': (struct_ants2fsl, 'fsl_matrix')},
            wall_time=5,
            requirements=[fsl_req.v('5.0.9')])

        struct2epi = pipeline.add(
            'struct2epi',
            ConvertXFM(
                invert_xfm=True),
            inputs={
                'in_file': (epi_ants2fsl, 'fsl_matrix')},
            wall_time=5,
            requirements=[fsl_req.v('5.0.9')])

        meanfunc = pipeline.add(
            'meanfunc',
            ImageMaths(op_string='-Tmean', suffix='_mean'),
            inputs={
                'in_file': ('preproc', nifti_gz_format)},
            wall_time=5,
            requirements=[fsl_req.v('5.0.9')])

        pipeline.add(
            'prep_fix',
            PrepareFIX(),
            inputs={
                'melodic_dir': ('melodic_ica', directory_format),
                't1_brain': ('coreg_ref_brain', nifti_gz_format),
                'mc_par': ('mc_par', par_format),
                'epi_brain_mask': ('brain_mask', nifti_gz_format),
                'epi_preproc': ('preproc', nifti_gz_format),
                'filtered_epi': ('filtered_data', nifti_gz_format),
                'epi2t1_mat': (epi_ants2fsl, 'fsl_matrix'),
                't12MNI_mat': (struct_ants2fsl, 'fsl_matrix'),
                'MNI2t1_mat': (MNI2t1, 'out_file'),
                't12epi_mat': (struct2epi, 'out_file'),
                'epi_mean': (meanfunc, 'out_file')},
            outputs={
                'fix_dir': ('fix_dir', directory_format),
                'hand_label_noise': ('hand_label_file', text_format)})

        return pipeline

    def fix_classification_pipeline(self, **name_maps):

        pipeline = self.new_pipeline(
            name='fix_classification',
            desc=("Automatic classification of noisy components from the "
                  "rsfMRI data using fsl FIX."),
            citations=[fsl_cite],
            name_maps=name_maps)

        pipeline.add(
            "fix",
            FSLFIX(
                component_threshold=self.parameter('component_threshold'),
                motion_reg=self.parameter('motion_reg'),
                classification=True),
            inputs={
                "feat_dir": ("fix_dir", directory_format),
                "train_data": ("train_data", rfile_format)},
            outputs={
                'labelled_components': ('label_file', text_format)},
            wall_time=30,
            requirements=[fsl_req.v('5.0.9'), fix_req.v('1.0')])

        return pipeline

    def fix_regression_pipeline(self, **name_maps):

        pipeline = self.new_pipeline(
            name='signal_regression',
            desc=("Regression of the noisy components from the rsfMRI data "
                  "using a python implementation equivalent to that in FIX."),
            citations=[fsl_cite],
            name_maps=name_maps)

        pipeline.add(
            "signal_reg",
            SignalRegression(
                motion_regression=self.parameter('motion_reg'),
                highpass=self.parameter('highpass')),
            inputs={
                "fix_dir": ("fix_dir", directory_format),
                "labelled_components": ("labelled_components", text_format)},
            outputs={
                'cleaned_file': ('output', nifti_gz_format)},
            wall_time=30,
            requirements=[fsl_req.v('5.0.9'), fix_req.v('1.0')])

        return pipeline

    def timeseries_normalization_to_atlas_pipeline(self, **name_maps):

        pipeline = self.new_pipeline(
            name='timeseries_normalization_to_atlas_pipeline',
            desc=("Apply ANTs transformation to the fmri filtered file to "
                  "normalize it to MNI 2mm."),
            citations=[fsl_cite],
            name_maps=name_maps)

        merge_trans = pipeline.add(
            'merge_transforms',
            NiPypeMerge(3),
            inputs={
                'in1': ('coreg_to_template_warp', nifti_gz_format),
                'in2': ('coreg_to_template_mat', text_matrix_format),
                'in3': ('coreg_matrix', text_matrix_format)},
            wall_time=1)

        pipeline.add(
            'ApplyTransform',
            ApplyTransforms(
                reference_image=self.parameter('MNI_template'),
                interpolation='Linear',
                input_image_type=3),
            inputs={
                'input_image': ('cleaned_file', nifti_gz_format),
                'transforms': (merge_trans, 'out')},
            outputs={
                'normalized_ts': ('output_image', nifti_gz_format)},
            wall_time=7,
            mem_gb=24,
            requirements=[ants_req.v('2')])

        return pipeline

    def smoothing_pipeline(self, **name_maps):

        pipeline = self.new_pipeline(
            name='smoothing_pipeline',
            desc=("Spatial smoothing of the normalized fmri file"),
            citations=[fsl_cite],
            name_maps=name_maps)

        pipeline.add(
            '3dBlurToFWHM',
            BlurToFWHM(
                fwhm=5,
                out_file='smoothed_ts.nii.gz',
                mask=self.parameter('MNI_template_mask')),
            inputs={
                'in_file': ('normalized_ts', nifti_gz_format)},
            outputs={
                'smoothed_ts': ('out_file', nifti_gz_format)},
            wall_time=5,
            requirements=[afni_req.v('16.2.10')])

        return pipeline


class MultiBoldMixin(MultiStudy):
    """
    A mixin class used for studies with an array of fMRI scans. Can be used to
    perform combined analysis over the array
    """

    add_data_specs = [
        FilesetSpec('train_data', rfile_format, 'fix_training_pipeline',
                    frequency='per_study'),
        FilesetSpec('fmri_pre-processeing_results', directory_format,
                    'gather_fmri_result_pipeline'),
        FilesetSpec('group_melodic', directory_format,
                    'group_melodic_pipeline')]

    @classmethod
    def fmri_substudies(cls):
        # Detect fMRI sub-studies
        fmri_substudies = []
        for substudy_spec in cls.substudy_specs():
            try:
                cls.data_spec(substudy_spec.inverse_map('fix_dir'))
            except ArcanaNameError:
                continue  # Sub study is not a Fmri study
            else:
                fmri_substudies.append(substudy_spec.name)

    def fix_training_pipeline(self, **name_maps):

        pipeline = self.new_pipeline(
            name='training_fix',
            desc=("Pipeline to create the training set for FIX given a group "
                  "of subjects with the hand_label_noise.txt file within "
                  "their fix_dir."),
            citations=[fsl_cite],
            name_maps=name_maps)

        num_fix_dirs = len(self.bold_substudies())
        merge_fix_dirs = pipeline.add(
            'merge_fix_dirs',
            NiPypeMerge(num_fix_dirs))
        merge_label_files = pipeline.add(
            'merge_label_files',
            NiPypeMerge(num_fix_dirs))
        for i, substudy_name in enumerate(self.bold_substudies(), start=1):
            spec = self.substudy_spec(substudy_name)
            pipeline.connect_input(
                spec.inverse_map('fix_dir'), merge_fix_dirs, 'in{}'.format(i),
                directory_format)
            pipeline.connect_input(
                spec.inverse_map('hand_label_noise'), merge_label_files,
                'in{}'.format(i), text_format)

        merge_visits = pipeline.add(
            IdentityInterface(
                ['list_dir', 'list_label_files']),
            inputs={
                'list_dir': (merge_fix_dirs, 'out'),
                'list_label_files': (merge_label_files, 'out')},
            joinsource=self.SUBJECT_ID,
            joinfield=['list_dir', 'list_label_files'], name='merge_visits')

        merge_subjects = pipeline.add(
            'merge_subjects',
            NiPypeMerge(
                2,
                ravel_inputs=True),
            inputs={
                'in1': (merge_visits, 'list_dir'),
                'in2': (merge_visits, 'list_label_files')},
            joinsource=self.SUBJECT_ID,
            joinfield=['in1', 'in2'])

        prepare_training = pipeline.add(
            'prepare_training',
            PrepareFIXTraining(
                epi_number=num_fix_dirs),
            inputs={
                'inputs_list': (merge_subjects, 'out')})

        pipeline.add(
            'fix_training',
            FSLFixTraining(
                outname='FIX_training_set',
                training=True),
            inputs={
                'list_dir': (prepare_training, 'prepared_dirs')},
            outputs={
                'train_data': ('training_set', rfile_format)},
            wall_time=240,
            requirements=[fix_req.v('1.0')])

        return pipeline

    def gather_fmri_result_pipeline(self, **name_maps):

        pipeline = self.new_pipeline(
            name='gather_fmri',
            desc=("Pipeline to gather together all the pre-processed "
                  "fMRI images"),
            version=1,
            citations=[fsl_cite],
            name_maps=name_maps)

        merge_inputs = pipeline.add(
            'merge_inputs',
            NiPypeMerge(len(self.bold_substudies())))
        for i, substudy_name in enumerate(self.bold_substudies(), start=1):
            spec = self.substudy_spec(substudy_name)
            pipeline.connect_input(
                spec.inverse_map('smoothed_ts'),
                merge_inputs, 'in{}'.format(i), nifti_gz_format)

        pipeline.add(
            'copy2dir',
            CopyToDir(),
            inputs={
                'in_files': (merge_inputs, 'out')},
            outputs={
                'fmri_pre-processeing_results': ('out_dir', directory_format)})

        return pipeline

    def group_melodic_pipeline(self, **name_maps):

        pipeline = self.new_pipeline(
            name='group_melodic',
            desc=("Group ICA"),
            citations=[fsl_cite],
            name_maps=name_maps)

        pipeline.add(
            MELODIC(
                no_bet=True,
                bg_threshold=self.parameter('brain_thresh_percent'),
                bg_image=self.parameter('MNI_template'),
                dim=self.parameter('group_ica_components'),
                report=True,
                out_stats=True,
                mm_thresh=0.5,
                sep_vn=True,
                mask=self.parameter('MNI_template_mask'),
                out_dir='group_melodic.ica'),
            inputs={
                'in_files': ('smoothed_ts', nifti_gz_format),
                'tr_sec': ('tr', float)},
            outputs={
                'group_melodic': ('out_dir', directory_format)},
            joinsource=self.SUBJECT_ID,
            joinfield=['in_files'],
            name='gica',
            requirements=[fsl_req.v('5.0.10')],
            wall_time=7200)

        return pipeline


def create_multi_fmri_class(name, t1, epis, epi_number, echo_spacing,
                            fm_mag=None, fm_phase=None, run_regression=False):

    inputs = []
    dct = {}
    data_specs = []
    parameter_specs = []
    output_files = []
    distortion_correction = False

    if fm_mag and fm_phase:
        logger.info(
            'Both magnitude and phase field map images provided. EPI '
            'ditortion correction will be performed.')
        distortion_correction = True
    elif fm_mag or fm_phase:
        logger.info(
            'In order to perform EPI ditortion correction both magnitude '
            'and phase field map images must be provided.')
    else:
        logger.info(
            'No field map image provided. Distortion correction will not be'
            'performed.')

    study_specs = [SubStudySpec('t1', T1Study)]
    ref_spec = {'t1_brain': 'coreg_ref_brain'}
    inputs.append(InputFileset('t1_primary', t1, dicom_format,
                                  is_regex=True, order=0))
    epi_refspec = ref_spec.copy()
    epi_refspec.update({'t1_wm_seg': 'coreg_ref_wmseg',
                        't1_preproc': 'coreg_ref',
                        'train_data': 'train_data'})
    study_specs.append(SubStudySpec('epi_0', BoldStudy, epi_refspec))
    if epi_number > 1:
        epi_refspec.update({'t1_wm_seg': 'coreg_ref_wmseg',
                            't1_preproc': 'coreg_ref',
                            'train_data': 'train_data',
                            'epi_0_coreg_to_tmpl_warp': 'coreg_to_tmpl_warp',
                            'epi_0_coreg_to_template_mat': 'coreg_to_template_mat'})
        study_specs.extend(SubStudySpec('epi_{}'.format(i), BoldStudy,
                                        epi_refspec)
                           for i in range(1, epi_number))

    study_specs.extend(SubStudySpec('epi_{}'.format(i), BoldStudy,
                                    epi_refspec)
                       for i in range(epi_number))

    for i in range(epi_number):
        inputs.append(InputFileset(
            'epi_{}_primary'.format(i), epis, dicom_format, order=i,
            is_regex=True))
#     inputs.extend(InputFileset(
#         'epi_{}_hand_label_noise'.format(i), text_format,
#         'hand_label_noise_{}'.format(i+1))
#         for i in range(epi_number))
        parameter_specs.append(
            ParamSpec('epi_{}_fugue_echo_spacing'.format(i), echo_spacing))

    if distortion_correction:
        inputs.extend(InputFileset(
            'epi_{}_field_map_mag'.format(i), fm_mag, dicom_format,
            dicom_tags={IMAGE_TYPE_TAG: MAG_IMAGE_TYPE}, is_regex=True,
            order=0)
            for i in range(epi_number))
        inputs.extend(InputFileset(
            'epi_{}_field_map_phase'.format(i), fm_phase, dicom_format,
            dicom_tags={IMAGE_TYPE_TAG: PHASE_IMAGE_TYPE}, is_regex=True,
            order=0)
            for i in range(epi_number))
    if run_regression:
        output_files.extend('epi_{}_smoothed_ts'.format(i)
                            for i in range(epi_number))
    else:
        output_files.extend('epi_{}_fix_dir'.format(i)
                            for i in range(epi_number))

    dct['add_substudy_specs'] = study_specs
    dct['add_data_specs'] = data_specs
    dct['add_param_specs'] = parameter_specs
    dct['__metaclass__'] = MultiStudyMetaClass
    return (MultiStudyMetaClass(name, (MultiBoldMixin,), dct), inputs,
            output_files)



from nipype.interfaces.base import (
    TraitedSpec, traits, File, CommandLineInputSpec, CommandLine)
import os
from nipype.interfaces.base import isdefined

ants_reg_path = os.path.abspath(
    os.path.join(os.path.dirname(__file__), 'resources', 'bash',
                 'antsRegistrationSyN.sh'))


class AntsRegSynInputSpec(CommandLineInputSpec):

    _trans_types = ['t', 'r', 'a', 's', 'sr', 'so', 'b', 'br', 'bo']
    _precision_types = ['f', 'd']
    input_file = File(mandatory=True, desc='existing input image',
                      argstr='-m %s', exists=True)
    ref_file = File(mandatory=True, desc='existing reference image',
                    argstr='-f %s', exists=True)
    num_dimensions = traits.Int(desc='number of dimension of the input file',
                                argstr='-d %s', mandatory=True)
    out_prefix = traits.Str(
        desc='A prefix that is prepended to all output files', argstr='-o %s',
        mandatory=True)
    transformation = traits.Enum(
        *_trans_types, argstr='-t %s',
        desc='type of transformation. t:translation, r:rigid, a:rigid+affine,'
        's:rigid+affine+deformable Syn, sr:rigid+deformable Syn, so:'
        'deformable Syn, b:rigid+affine+deformable b-spline Syn, br:'
        'rigid+deformable b-spline Syn, bo:deformable b-spline Syn')
    num_threads = traits.Int(desc='number of threads', argstr='-n %s')
    radius = traits.Float(
        desc='radius for cross correlation metric used during SyN stage'
        ' (default = 4)', argstr='-r %f')
    spline_dist = traits.Float(
        desc='spline distance for deformable B-spline SyN transform'
        ' (default = 26)', argstr='-s %f')
    ref_mask = File(
        desc='mask for the fixed image space', exists=True, argstr='-x %s')
    precision_type = traits.List(
        traits.Enum(*_precision_types), argstr='-p %s', desc='precision type '
        '(default = d). f:float, d:double')
    use_histo_match = traits.Int(desc='use histogram matching (default = 0).'
                                 '0: False, 1:True', argstr='-j %s')


class AntsRegSynOutputSpec(TraitedSpec):
    regmat = File(desc="Linear transformation matrix")
    reg_file = File(desc="Registered image")
    warp_file = File(desc="non-linear warp file")
    inv_warp = File(desc='invert of the warp file')


class AntsRegSyn(CommandLine):

    _cmd = ants_reg_path
    input_spec = AntsRegSynInputSpec
    output_spec = AntsRegSynOutputSpec
    mat_ext = '.mat'
    img_ext = '.nii.gz'

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs['regmat'] = os.path.join(os.getcwd(),
                                         self._gen_filename('regmat'))
        outputs['reg_file'] = os.path.join(os.getcwd(),
                                           self._gen_filename('reg_file'))
        if isdefined(self.inputs.transformation and
                     (self.inputs.transformation != 'r' or
                      self.inputs.transformation != 'a' or
                      self.inputs.transformation != 't')):
            outputs['warp_file'] = os.path.join(
                os.getcwd(), self._gen_filename('warp_file'))
            outputs['inv_warp'] = os.path.join(
                os.getcwd(), self._gen_filename('inv_warp'))

        return outputs

    def _gen_filename(self, name):
        if name == 'regmat':
            fid = os.path.basename(self.inputs.out_prefix)
            fname = fid + '_0GenericAffine' + self.mat_ext
        elif name == 'warp_file':
            fid = os.path.basename(self.inputs.out_prefix)
            fname = fid + '_1Warp' + self.img_ext
        elif name == 'inv_warp':
            fid = os.path.basename(self.inputs.out_prefix)
            fname = fid + '_1InverseWarp' + self.img_ext
        elif name == 'reg_file':
            fid = os.path.basename(self.inputs.out_prefix)
            fname = fid + self.img_ext
        else:
            assert False
        return fname
